@@ -34,8 +34,12 @@ static int _buf_flush(Buffer_t *Buffer)
 {
 	int ret;
 
-	if (!Buffer->Next || Buffer->current < 0L || !Buffer->dirty)
+	if (!Buffer->Next || !Buffer->dirty)
 		return 0;
+	if(Buffer->current < 0L) {
+		fprintf(stderr,"Should not happen\n");
+		return 0;
+	}
 	ret = force_write(Buffer->Next, 
 			  Buffer->buf + Buffer->dirty_pos,
 			  Buffer->current + Buffer->dirty_pos,
@@ -169,9 +173,6 @@ Stream_t *buf_init(Stream_t *Next, int size, int grain, int grain2)
 	Buffer_t *Buffer;
 	Stream_t *Stream;
 
-#if 0
-	return Next;
-#else
 
 	if(Next->Buffer){
 		Next->refs--;
@@ -198,6 +199,5 @@ Stream_t *buf_init(Stream_t *Next, int size, int grain, int grain2)
 	Buffer->Buffer = 0;
 	Buffer->Next->Buffer = (Stream_t *) Buffer;
 	return Stream;
-#endif
 }
 
