@@ -20,14 +20,21 @@ extern int errno;
 
 unsigned int num_clus;			/* total number of cluster */
 
+#ifndef linux
+#define BOOTSIZE 512
+#else
+#define BOOTSIZE 256
+#endif
+
+
 /*
  * Read the boot sector.  We glean the disk parameters from this sector.
  */
 static int read_boot(Stream_t *Stream, struct bootsector * boot)
 {	
 	/* read the first sector, or part of it */
-	if (force_read(Stream, (char *) boot, 0, sizeof(struct bootsector)) !=
-	    sizeof(struct bootsector))
+	if (force_read(Stream, (char *) boot, 0, BOOTSIZE) !=
+	    BOOTSIZE)
 		return -1;
 	return 0;
 }
@@ -68,6 +75,12 @@ static int get_media_type(Stream_t *St, struct bootsector *boot)
 }
 
 
+Stream_t *GetFs(Stream_t *Fs)
+{
+	while(Fs && Fs->Class != &FsClass)
+		Fs = Fs->Next;
+	return Fs;
+}
 
 Stream_t *find_device(char drive, int mode, struct device *out_dev,
 		      struct bootsector *boot,
@@ -93,10 +106,10 @@ Stream_t *find_device(char drive, int mode, struct device *out_dev,
 		Stream = 0;
 #endif
 
-		if (!Stream){
-			Stream = SimpleFloppyOpen(out_dev, dev, name, mode,
-						  errmsg);
-		} else
+		if (!Stream)
+			Stream = SimpleFileOpen(out_dev, dev, name, mode,
+						errmsg, 0);
+		else
 			out_dev->use_2m = 0x7f;
 
 		if( !Stream)
@@ -104,7 +117,9 @@ Stream_t *find_device(char drive, int mode, struct device *out_dev,
 
 		/* read the boot sector */
 		if (read_boot(Stream, boot)){
-			sprintf(errmsg,"init %c: read boot", drive);
+			sprintf(errmsg,
+				"init %c: could not read boot sector",
+				drive);
 			continue;
 		}
 
@@ -114,6 +129,7 @@ Stream_t *find_device(char drive, int mode, struct device *out_dev,
 		}
 
 		/* set new parameters, if needed */
+		errno = 0;
 		if(SET_GEOM(Stream, out_dev, dev, *media, boot)){
 			sprintf(errmsg, "Can't set disk parameters for %c: %s", 
 				drive, strerror(errno));
@@ -185,16 +201,18 @@ Stream_t *fs_init(char drive, int mode)
 		 * all numbers are in sectors, except num_clus 
 		 * (which is in clusters)
 		 */
-		if(boot0.dos4 == 0x29) {
-			This->serialized = 1;
-			tot_sectors = DWORD(bigsect);
-			SET_INT(tot_sectors, WORD(psect));
-			This->serial_number = DWORD(serial);
+		tot_sectors = WORD(psect);
+		if(!tot_sectors) {
+			tot_sectors = DWORD(bigsect);			
 			nhs = DWORD(nhs);
-		} else {			
-			tot_sectors = WORD(psect);
+		} else
 			nhs = WORD(nhs);
+
+		if(boot0.ext.old.dos4 == 0x29 && WORD(fatlen)) {
+			This->serialized = 1;
+			This->serial_number = DWORD(ext.old.serial);
 		}
+
 		This->cluster_size = boot0.clsiz; 		
 		This->fat_start = WORD(nrsvsect);
 		This->fat_len = WORD(fatlen);
@@ -202,27 +220,12 @@ Stream_t *fs_init(char drive, int mode)
 		This->num_fat = boot0.nfat;
 	}	
 	cylinder_size = dev.sectors * dev.heads;
-	This->dir_start = This->fat_start + This->num_fat * This->fat_len;
 
-	if (!tot_sectors ){
-		/* this should only happen on hard disks, so let's
-		 * assume a 16 bit fat for now */
-		tot_sectors = This->dir_start + This->dir_len + 
-			This->fat_len * This->sector_size / 
-				2 * This->cluster_size;
-		tot_sectors -= (tot_sectors + nhs) % cylinder_size;
-		fprintf(stderr,
-			"Warning: disk size unknown, assuming %d Megabytes\n", 
-			(int) (tot_sectors * This->sector_size / 1024 / 1024));
-	}
-
-	This->num_clus = (tot_sectors - This->dir_start - This->dir_len) /
-		This->cluster_size;
 	if(!mtools_skip_check && (tot_sectors % dev.sectors)){
 		fprintf(stderr,
 			"Total number of sectors not a multiple of"
 			" sectors per track!\n");
-		cleanup_and_exit(1);
+		exit(1);
 	}
 
 	/* full cylinder buffering */
@@ -242,6 +245,14 @@ Stream_t *fs_init(char drive, int mode)
 #ifdef linux
 	disk_size = cylinder_size;
 #endif
+
+#if 1
+	if(disk_size > 256) {
+		disk_size = dev.sectors;
+		if(dev.sectors % 2)
+			disk_size <<= 1;
+	}
+#endif
 	if (disk_size % 2)
 		disk_size *= 2;
 
@@ -258,32 +269,12 @@ Stream_t *fs_init(char drive, int mode)
 		This->Next = This->Direct;
 	}
 
-	if (dev.use_2m & 0x7f)
-		This->num_fat = 1;
 	/* read the FAT sectors */
-	if(fat_read(This, &boot0, dev.fat_bits)){
+	if(fat_read(This, &boot0, dev.fat_bits, tot_sectors, dev.use_2m&0x7f)){
+		This->num_fat = 1;
 		FREE(&This->Next);
 		Free(This->Next);
 		return NULL;
 	}
 	return (Stream_t *) This;
 }
-
-/*
- * Get the amount of free space on the diskette
- */
-
-long getfree(Stream_t *Stream)
-{
-	DeclareThis(Fs_t);
-	register unsigned int i;
-	size_t total;
-
-	total = 0L;
-	for (i = 2; i < This->num_clus + 2; i++) {
-		/* if fat_decode returns zero */
-		if (!This->fat_decode(This,i))
-			total += This->cluster_size;
-	}
-	return(total * This->sector_size);
-}
