@@ -46,11 +46,11 @@ static int parse_range(const char **p, const char *s, char *out,
 			for(i=first; i<last; i++)
 				table[i] = 1;
 		} else
-			table[*((*p)++)] = 1;
+			table[(int) *((*p)++)] = 1;
 	}
 	if(out)
 		*out = *s;
-	if(table[*s])
+	if(table[(int) *s])
 		return 1 ^ reverse;
 	if(compfn == exactcmp)
 		return reverse;
@@ -68,11 +68,9 @@ static int parse_range(const char **p, const char *s, char *out,
 }
 
 
-static int _match(__const char *s, __const char *p, char *out, int Case,
+static int _match(const char *s, const char *p, char *out, int Case,
 		  int (*compfn) (char a, char b))
 {
-	int matched, reverse;
-
 	for (; *p != '\0'; ) {
 		switch (*p) {
 			case '?':	/* match any one character */
@@ -123,7 +121,7 @@ static int _match(__const char *s, __const char *p, char *out, int Case,
 }
 
 
-int match(__const char *s, __const char *p, char *out, int Case)
+int match(const char *s, const char *p, char *out, int Case)
 {
 	int (*compfn)(char a, char b);
 
@@ -134,3 +132,8 @@ int match(__const char *s, __const char *p, char *out, int Case)
 		compfn = casecmp;
 	return _match(s, p, out, Case, compfn);
 }
+
+int hasWildcards(const char *string)
+{
+	return (int) strpbrk(string, "?*");
+}
