@@ -11,7 +11,7 @@
 #include "msdos.h"
 #include "mtools.h"
 #include "vfat.h"
-#include "streamcache.h"
+#include "mainloop.h"
 #include "plain_io.h"
 #include "nameclash.h"
 #include "file.h"
@@ -21,7 +21,7 @@
  * Preserve the file modification times after the fclose()
  */
 
-static void set_mtime(char *target, long mtime)
+static void set_mtime(const char *target, long mtime)
 {
 #ifdef HAVE_UTIMES
 	struct timeval tv[2];
@@ -54,23 +54,20 @@ typedef struct Arg_t {
 	int needfilter;
 	int nowarn;
 	int single;
-	int interactive;
 	int verbose;
 	int type;
-	StreamCache_t sc;
+	MainParam_t mp;
 	ClashHandling_t ch;
 	Stream_t *sourcefile;
 	Stream_t *targetDir;
-	Stream_t *targetFs;
-	char *progname;
 } Arg_t;
 
-static int read_file(Stream_t *Dir, StreamCache_t *sc, int entry, 
+static int read_file(Stream_t *Dir, MainParam_t *mp, int entry, 
 		     int needfilter)
 {
-	Arg_t *arg=(Arg_t *) sc->arg;
+	Arg_t *arg=(Arg_t *) mp->arg;
 	long mtime;
-	Stream_t *File=sc->File;
+	Stream_t *File=mp->File;
 	Stream_t *Target, *Source;
 	struct stat stbuf;
 	int ret;
@@ -82,7 +79,7 @@ static int read_file(Stream_t *Dir, StreamCache_t *sc, int entry,
 		mtime = 0L;
 
 	/* if we are creating a file, check whether it already exists */
-	if (!arg->nowarn && strcmp(arg->target, "-") && 
+	if (!arg->nowarn && arg->target && strcmp(arg->target, "-") && 
 	    !access(arg->target, 0)){
 		if( ask_confirmation("File \"%s\" exists, overwrite (y/n) ? ",
 				     arg->target,0))
@@ -97,12 +94,12 @@ static int read_file(Stream_t *Dir, StreamCache_t *sc, int entry,
 	}
 
 	if(!arg->type)
-		fprintf(stderr,"Copying %s\n", sc->outname);
+		fprintf(stderr,"Copying %s\n", mp->outname);
 	if(got_signal)
 		return MISSED_ONE;
-	if ((Target = SimpleFloppyOpen(0, 0, arg->target,
-				       O_WRONLY | O_CREAT | O_TRUNC,
-				       errmsg))) {
+	if ((Target = SimpleFileOpen(0, 0, arg->target,
+				     O_WRONLY | O_CREAT | O_TRUNC,
+				     errmsg, 0))) {
 		ret = 0;
 		if(needfilter && arg->textmode){
 			Source = open_filter(COPY(File));
@@ -119,7 +116,7 @@ static int read_file(Stream_t *Dir, StreamCache_t *sc, int entry,
 			unlink(arg->target);
 			return MISSED_ONE;
 		}
-		if(strcmp(arg->target,"-"))
+		if(arg->target && strcmp(arg->target,"-"))
 			set_mtime(arg->target, mtime);
 		return GOT_ONE;
 	} else {
@@ -128,15 +125,15 @@ static int read_file(Stream_t *Dir, StreamCache_t *sc, int entry,
 	}
 }
 
-static  int dos_read(Stream_t *Dir, StreamCache_t *sc, int entry)
+static  int dos_read(Stream_t *Dir, MainParam_t *mp, int entry)
 {
-	return read_file(Dir, sc, entry, 1);
+	return read_file(Dir, mp, entry, 1);
 }
 
 
-static  int unix_read(char *name, StreamCache_t *sc)
+static  int unix_read(char *name, MainParam_t *mp)
 {
-	return read_file(0, sc, -1, 0);
+	return read_file(0, mp, -1, 0);
 }
 
 
@@ -150,13 +147,13 @@ static struct directory *writeit(char *dosname,
 				 struct directory *dir)
 {
 	Stream_t *Target;
-	long now;
+	time_t now;
 	int type, fat, ret;
 	long date;
 	size_t filesize, newsize;
 	Arg_t *arg = (Arg_t *) arg0;
 
-	if (arg->sc.File->Class->get_data(arg->sc.File,
+	if (arg->mp.File->Class->get_data(arg->mp.File,
 					  & date, &filesize, &type, 0) < 0 ){
 		fprintf(stderr, "Can't stat source file\n");
 		return NULL;
@@ -174,7 +171,7 @@ static struct directory *writeit(char *dosname,
 		return NULL;
 
 	/* will it fit? */
-	if (filesize > getfree(arg->targetFs)) {
+	if (!getfreeMin(arg->targetDir, filesize)) {
 		fprintf(stderr,"Disk full\n");
 		return NULL;
 	}
@@ -186,18 +183,18 @@ static struct directory *writeit(char *dosname,
 		time(&now);
 
 	mk_entry(dosname, 0x20, 0, 0, now, dir);
-	Target = open_file(COPY(arg->targetFs), dir);
+	Target = open_file(arg->targetDir, dir);
 	if(!Target){
 		fprintf(stderr,"Could not open Target\n");
-		cleanup_and_exit(1);
+		exit(1);
 	}
 	if (arg->needfilter & arg->textmode)
 		Target = open_filter(Target);
-	ret = copyfile(arg->sc.File, Target);
+	ret = copyfile(arg->mp.File, Target);
 	GET_DATA(Target, 0, &newsize, 0, &fat);
 	FREE(&Target);
 	if(ret < 0 ){
-		fat_free(arg->targetFs, fat);
+		fat_free(arg->targetDir, fat);
 		return NULL;
 	} else {
 		mk_entry(dosname, 0x20, fat, newsize, now, dir);
@@ -206,12 +203,12 @@ static struct directory *writeit(char *dosname,
 }
 
 
-static int write_file(Stream_t *Dir, StreamCache_t *sc, int entry,
+static int write_file(Stream_t *Dir, MainParam_t *mp, int entry,
 		      int needfilter)
 /* write a messy dos file to another messy dos file */
 {
 	int result;
-	Arg_t * arg = (Arg_t *) (sc->arg);
+	Arg_t * arg = (Arg_t *) (mp->arg);
 
 	arg->needfilter = needfilter;
 	if (arg->targetDir == Dir){
@@ -221,8 +218,7 @@ static int write_file(Stream_t *Dir, StreamCache_t *sc, int entry,
 		arg->ch.ignore_entry = -1;
 		arg->ch.source = -2;
 	}
-	result = mwrite_one(arg->targetDir, arg->targetFs,
-			    arg->progname, arg->target, 0,
+	result = mwrite_one(arg->targetDir, arg->target, 0,
 			    writeit, (void *)arg, &arg->ch);
 	if(result == 1)
 		return GOT_ONE;
@@ -231,51 +227,54 @@ static int write_file(Stream_t *Dir, StreamCache_t *sc, int entry,
 }
 
 
-static int dos_write(Stream_t *Dir, StreamCache_t *sc, int entry)
+static int dos_write(Stream_t *Dir, MainParam_t *mp, int entry)
 {
-	return write_file(Dir, sc, entry, 0);
+	return write_file(Dir, mp, entry, 0);
 }
 
-static int unix_write(char *name, StreamCache_t *sc)
+static int unix_write(char *name, MainParam_t *mp)
 /* write a Unix file to a messy DOS fs */
 {
-	return write_file(0, sc, 0, 1);
+	return write_file(0, mp, 0, 1);
 }
 
 
+static void usage(void)
+{
+	fprintf(stderr,
+		"Mtools version %s, dated %s\n", mversion, mdate);
+	fprintf(stderr,
+		"Usage: %s [-tnmvV] sourcefile targetfile\n", progname);
+	fprintf(stderr,
+		"       %s [-tnmvV] sourcefile [sourcefiles...] targetdirectory\n", 
+		progname);
+	exit(1);
+}
+
 void mcopy(int argc, char **argv, int mtype)
 {
-	Stream_t *SubDir,*Dir, *Fs;
+	Stream_t *SubDir,*Dir;
 	int have_target;
 	Arg_t arg;
-	int c, oops, ret;
+	int c, ret, fastquit;
 	char filename[VBUFSIZE];
+	char spareOutname[VBUFSIZE];
 	
 	struct stat stbuf;
 
-	int interactive;
-
-	arg.progname = argv[0];
-
 	/* get command line options */
 
 	init_clash_handling(& arg.ch);
 
-	oops = interactive = 0;
-
-
 	/* get command line options */
 	arg.preserve = 0;
 	arg.nowarn = 0;
 	arg.textmode = 0;
-	arg.interactive = 0;
 	arg.verbose = 0;
 	arg.type = mtype;
-	while ((c = getopt(argc, argv, "itnmvorsamORSAM")) != EOF) {
+	fastquit = 0;
+	while ((c = getopt(argc, argv, "tnmvorsamQORSAM")) != EOF) {
 		switch (c) {
-			case 'i':
-				arg.interactive = 1;
-				break;
 			case 't':
 				arg.textmode = 1;
 				break;
@@ -288,33 +287,28 @@ void mcopy(int argc, char **argv, int mtype)
 			case 'v':	/* dummy option for mcopy */
 				arg.verbose = 1;
 				break;
-			case '?':
-				oops = 1;
+			case 'Q':
+				fastquit = 1;
 				break;
+			case '?':
+				usage();
 			default:
 				if(handle_clash_options(&arg.ch, c))
-					oops=1;
+					usage();
 				break;
 		}
 	}
 
-	if (oops || (argc - optind) < 1) {
-		fprintf(stderr,
-			"Mtools version %s, dated %s\n", mversion, mdate);
-		fprintf(stderr,
-			"Usage: %s [-itnmvV] sourcefile targetfile\n", argv[0]);
-		fprintf(stderr,
-			"       %s [-itnmvV] sourcefile [sourcefiles...] targetdirectory\n", 
-			argv[0]);
-		cleanup_and_exit(1);
-	}
+	if (argc - optind < 1)
+		usage();
 
 	/* only 1 file to copy... */
 	arg.single = SINGLE;
 	
-	init_sc(&arg.sc);		
-	arg.sc.arg = (void *) &arg;
-	arg.sc.openflags = O_RDONLY;
+	init_mp(&arg.mp);
+	arg.mp.fast_quit = fastquit;
+	arg.mp.arg = (void *) &arg;
+	arg.mp.openflags = O_RDONLY;
 
 	arg.targetDir = NULL;
 	if(mtype){
@@ -322,84 +316,87 @@ void mcopy(int argc, char **argv, int mtype)
 		/* Mtype = copying to stdout */
 
 		have_target = 0;
-		arg.target = "-";
-		arg.sc.outname = filename;
+		arg.target = 0;
+		arg.mp.outname = filename;
 		arg.single = 0;		
-		arg.sc.callback = dos_read;
-		arg.sc.unixcallback = unix_read;
+		arg.mp.callback = dos_read;
+		arg.mp.unixcallback = unix_read;
 	} else if(argc - optind >= 2 && 
 		  argv[argc-1][0] && argv[argc-1][1] == ':'){
 
 		/* Copying to Dos */
-
+		
 		have_target = 1;
-		Dir = open_subdir(&arg.sc, argv[argc-1], O_RDWR, &Fs);
-		arg.targetFs = Fs;
+		Dir = open_subdir(&arg.mp, argv[argc-1], O_RDWR, 0);
 		if(!Dir){
 			fprintf(stderr,"Bad target\n");
-			cleanup_and_exit(1);
+			exit(1);
 		}
-		get_name(argv[argc-1], filename, arg.sc.mcwd);
-		SubDir = descend(Dir, Fs, filename, 0, 0);
+		get_name(argv[argc-1], filename, arg.mp.mcwd);
+		SubDir = descend(Dir, filename, 0, 0);
 		if (!SubDir){
 			/* the last parameter is a file */
-			if ( argc - optind != 2 ){
+			if (argc - optind != 2){
 				fprintf(stderr,
-					"%s: Too many arguments, or destination directory omitted\n", argv[0]);
-				cleanup_and_exit(1);
+					"%s: Too many arguments, or destination directory omitted\n",
+					argv[0]);
+				FREE(&Dir);
+				exit(1);
 			}
 			arg.targetDir = Dir;
-			arg.sc.outname = 0; /* toss away source name */
+			arg.mp.outname = 0; /* toss away source name */
 			arg.target = filename; /* store near name given as
 						* target */
 			arg.single = 1;
 		} else {
 			arg.targetDir = SubDir;
 			FREE(&Dir);
-			arg.sc.outname = arg.target = filename;
+			arg.mp.outname = filename;
+			arg.target = filename;
 			arg.single = 0;
 		}
-		arg.sc.callback = dos_write;
-		arg.sc.unixcallback = unix_write;
+		arg.mp.callback = dos_write;
+		arg.mp.unixcallback = unix_write;
 	} else {
-
 		/* Copying to Unix */
-
 		have_target = 1;
 		if (argc - optind == 1){
 			/* one argument: copying to current directory */
 			arg.single = 0;
 			have_target = 0;
-			arg.target = arg.sc.outname = filename;
+			arg.target = filename;
+			arg.mp.outname = filename;
 		} else if (strcmp(argv[argc-1],"-") && 
 			   !stat(argv[argc-1], &stbuf) &&
 			   S_ISDIR(stbuf.st_mode)){
-			arg.target = (char *) safe_malloc(VBUFSIZE + 1 + 
-							  strlen(argv[argc-1]));
-			strcpy(arg.target, argv[argc-1]);
-			strcat(arg.target,"/");
-			arg.sc.outname = arg.target+strlen(arg.target);
+			char *tmp;
+			tmp = (char *) safe_malloc(VBUFSIZE + 1 + 
+						   strlen(argv[argc-1]));
+			strcpy(tmp, argv[argc-1]);
+			strcat(tmp,"/");
+			arg.mp.outname = tmp+strlen(tmp);
+			arg.target = tmp;
 		} else if (argc - optind != 2) {
 			/* last argument is not a directory: we should only
 			 * have one source file */
 			fprintf(stderr,
 				"%s: Too many arguments or destination directory omitted\n", 
 				argv[0]);
-			cleanup_and_exit(1);
+			exit(1);
 		} else {
 			arg.target = argv[argc-1];
-			arg.sc.outname = filename;
+			arg.mp.outname = filename;
 		}
 
-		arg.sc.callback = dos_read;
-		arg.sc.unixcallback = unix_read;
+		arg.mp.callback = dos_read;
+		arg.mp.unixcallback = unix_read;
 	}
+	if(!arg.mp.outname)
+		arg.mp.outname = spareOutname;
 
-	arg.sc.lookupflags = ACCEPT_PLAIN | DO_OPEN | arg.single;
-	ret=main_loop(&arg.sc, argv[0], argv + optind, 
-		      argc - optind - have_target);
+	arg.mp.lookupflags = ACCEPT_PLAIN | DO_OPEN | NO_DOTS | arg.single;
+	ret=main_loop(&arg.mp, argv + optind, argc - optind - have_target);
 	FREE(&arg.targetDir);
-	finish_sc(&arg.sc);
 
-	cleanup_and_exit(ret);
+	exit(ret);
 }
