@@ -4,7 +4,7 @@
  * written by:
  *
  * Alain L. Knaff			
- * Alain.Knaff@inrialpes.fr
+ * Alain.Knaff@poboxes.com
  *
  */
 
@@ -16,7 +16,7 @@
 #include "patchlevel.h"
 #include <linux/fd.h>
 #include <sys/wait.h>
-#include "streamcache.h"
+#include "mainloop.h"
 #include "fs.h"
 
 extern int errno;
@@ -34,41 +34,47 @@ void mmount(int argc, char **argv, int type)
 	
 	if (argc<2 || !argv[1][0]) {
 		fprintf(stderr,"Usage: %s -V drive:\n", argv[0]);
-		cleanup_and_exit(1);
+		exit(1);
 	}
 	drive = argv[1][0];
 	Stream = find_device(drive, O_RDONLY, &dev, &boot, name, &media);
 	if(!Stream)
-		cleanup_and_exit(1);
+		exit(1);
 	FREE(&Stream);
 
+	destroy_privs();
+
 	/* and finally mount it */
 	switch((pid=fork())){
 	case -1:
 		fprintf(stderr,"fork failed\n");
-		cleanup_and_exit(1);
+		exit(1);
 	case 0:
 		close(2);
 		open("/dev/null", O_RDWR);
-		argv[1] = "mount" ;
+		argv[1] = strdup("mount");
 		if ( argc > 2 )
 			execvp("mount", argv + 1 );
 		else
 			execlp("mount", "mount", name, 0);
 		perror("exec mount");
-		cleanup_and_exit(1);
+		exit(1);
 	default:
 		while ( wait(&status) != pid );
 	}	
 	if ( WEXITSTATUS(status) == 0 )
-		cleanup_and_exit(0);
-	argv[0] = "mount";
-	argv[1] = "-r";
+		exit(0);
+	argv[0] = strdup("mount");
+	argv[1] = strdup("-r");
+	if(!argv[0] || !argv[1]){
+		fprintf(stderr, "Out of memory error\n");
+		exit(1);
+	}
 	if ( argc > 2 )
 		execvp("mount", argv);
 	else
 		execlp("mount", "mount","-r", name, 0);
-	cleanup_and_exit(1);
+	exit(1);
 }
 
 #else /* linux */
@@ -78,7 +84,7 @@ void mmount(int argc, char **argv, int type)
 void mmount(int argc, char **argv, int type)
 {
   fprintf(stderr,"This command is only available for LINUX \n");
-  cleanup_and_exit(1);
+  exit(1);
 }
 #endif /* linux */
 
