@@ -2,13 +2,15 @@
 #include "msdos.h"
 #include "mtools.h"
 #include "patchlevel.h"
+#include "partition.h"
+#include "vfat.h"
 
 const char *mversion = VERSION;
 const char *mdate = DATE;
-int	mtools_raw_tty = 1;
+const char *progname;
 
 static const struct dispatch {
-	char *cmd;
+	const char *cmd;
 	void (*fn)(int, char **, int);
 	int type;
 } dispatch[] = {
@@ -20,34 +22,65 @@ static const struct dispatch {
 	{"mdeltree",mdel, 2},
 	{"mdir",mdir, 0},
 	{"mformat",mformat, 0},
+	{"minfo", minfo, 0},
 	{"mlabel",mlabel, 0},
 	{"mmd",mmd, 0},
 	{"mmount",mmount, 0},
+	{"mpartition",mpartition, 0},
 	{"mrd",mdel, 1},
 	{"mread",mcopy, 0},
 	{"mmove",mmove, 0},
 	{"mren",mmove, 1},
-	{"mtest", mtest, 0},
+	{"mtoolstest", mtoolstest, 0},
 	{"mtype",mcopy, 1},
 	{"mwrite",mcopy, 0},
+	{"mzip", mzip, 0}
 };
 #define NDISPATCH (sizeof dispatch / sizeof dispatch[0])
 
 
+
 void main(int argc,char **argv)
 {
 	char *name;
 	int i;
 
+	init_privs();
+
 	/* print the version */
 	if(argc >= 2 && strcmp(argv[1], "-V") == 0) {
 		printf("Mtools version %s, dated %s\n", mversion, mdate);
 		exit(0);
 	}
 
+	/* check whether the compiler lays out structures in a sane way */
+	if(sizeof(struct partition) != 16 ||
+	   sizeof(struct directory) != 32 ||
+	   sizeof(struct vfat_subentry) !=32) {
+		fprintf(stderr,"Mtools has not been correctly compiled\n");
+		fprintf(stderr,"Recompile it using a more recent compiler\n");
+		exit(137);
+	}
+
 	if ((name = strrchr(argv[0],'/')))
 		name++;
-	else name = argv[0];
+	else
+		name = argv[0];
+	progname = argv[0];
+
+	/* this allows the different tools to be called as "mtools -c <command>"
+	** where <command> is mdir, mdel, mcopy etcetera
+	** Mainly done for the BeOS, which doesn't support links yet.
+	*/
+
+	if(argc >= 3 && 
+	   !strcmp(argv[1], "-c") &&
+	   !strcmp(name, "mtools")) {
+		argc-=2;
+		argv+=2;
+		name = argv[0];
+	}
+
 	argv[0] = name;
 	
 	read_config();
@@ -66,5 +99,5 @@ void main(int argc,char **argv)
 	}
 	putc('\n', stderr);
 
-	cleanup_and_exit(1);
+	exit(1);
 }
