@@ -11,8 +11,9 @@
 extern int lockf(int, int, off_t);  /* SCO has no proper include file for lockf */
 #endif 
 
+
 typedef struct device {
-	char *name;	       	/* full path to device */
+	const char *name;       /* full path to device */
 
 	char drive;	       	/* the drive letter */
 	int fat_bits;		/* FAT encoding scheme */
@@ -23,20 +24,33 @@ typedef struct device {
 	unsigned int sectors;	/* sectors */
 
 	off_t offset;	       	/* skip this many bytes */
+
+	int partition;
+	int scsi;
+
+	int privileged;
+
+	/* rarely used stuff */
+	int nolock;	
+
+	/* Linux only stuff */
+	unsigned int ssize;
+	unsigned int use_2m;
 	int use_xdf;
 
+	char *precmd;		/* command to be executed before opening
+				 * the drive */
+
 	/* internal variables */
 	int file_nr;		/* used during parsing */
-	unsigned int ssize;
-	unsigned int use_2m;
-
-	int partition;
+	int hidden;		/* number of hidden sectors. Used for
+				 * mformatting partitioned devices */
 } device_t;
 
 #include "stream.h"
 
 
-extern char *short_illegals, *long_illegals;
+extern const char *short_illegals, *long_illegals;
 
 size_t maximize(size_t *target, long max);
 size_t minimize(size_t *target, long max);
@@ -54,27 +68,26 @@ int readwrite_sectors(int fd, /* file descriptor */
 		      int direction,
 		      int retries);
 
-int lock_dev(int fd, int mode);
-
-Stream_t *subdir(Stream_t *, char *pathname);
+int lock_dev(int fd, int mode, struct device *dev);
 
 char *unix_normalize (char *ans, char *name, char *ext);
 void dir_write(Stream_t *Dir, int num, struct directory *dir);
 char *dos_name(char *filename, int verbose, int *mangled, char *buffer);
-struct directory *mk_entry(char *filename, char attr,
+struct directory *mk_entry(const char *filename, char attr,
 			   unsigned int fat, size_t size, long date,
 			   struct directory *ndir);
 int copyfile(Stream_t *Source, Stream_t *Target);
-long getfree(Stream_t *Stream);
+unsigned long getfree(Stream_t *Stream);
+unsigned long getfreeMin(Stream_t *Stream, size_t ref);
 
 FILE *opentty(int mode);
 
 int is_dir(Stream_t *Dir, char *path);
-Stream_t *descend(Stream_t *Dir, Stream_t *Fs, char *path, int barf,
-		  char *outname);
+Stream_t *descend(Stream_t *Dir, char *path, int barf,char *outname);
 
-int dir_grow(Stream_t *Dir, Stream_t *Fs, int size);
-int match(__const char *, __const char *, char *, int);
+int dir_grow(Stream_t *Dir, int size);
+int match(const char *, const char *, char *, int);
+int hasWildcards(const char *string);
 
 Stream_t *new_file(char *filename, char attr,
 		   unsigned int fat, size_t size, long date,
@@ -91,13 +104,12 @@ void setup_signal(void);
 if(source)target=source
 
 
-UNUSED(static inline int compare (long ref, long testee));
-static inline int compare (long ref, long testee)
+UNUSED(static inline int compare (long ref, long testee))
 {
 	return (ref && ref != testee);
 }
 
-
+Stream_t *GetFs(Stream_t *Fs);
 Stream_t *find_device(char drive, int mode, struct device *out_dev,
 		      struct bootsector *boot,
 		      char *name, int *media);
@@ -110,7 +122,6 @@ struct directory *labelit(char *dosname,
 char *label_name(char *filename, int verbose, 
 		 int *mangled, char *ans);
 
-void cleanup_and_exit(int code);
 /* environmental variables */
 extern int mtools_skip_check;
 extern int mtools_fat_compatibility;
@@ -120,7 +131,7 @@ extern int mtools_rate_0, mtools_rate_any, mtools_raw_tty;
 void read_config(void);
 extern struct device *devices;
 extern struct device const_devices[];
-extern const nr_const_devices;
+extern const int nr_const_devices;
 
 #define New(type) ((type*)(malloc(sizeof(type))))
 #define Grow(adr,n,type) ((type*)(realloc((char *)adr,n*sizeof(type))))
@@ -134,10 +145,25 @@ void mcopy(int argc, char **argv, int type);
 void mdel(int argc, char **argv, int type);
 void mdir(int argc, char **argv, int type);
 void mformat(int argc, char **argv, int type);
+void minfo(int argc, char **argv, int type);
 void mlabel(int argc, char **argv, int type);
 void mmd(int argc, char **argv, int type);
 void mmount(int argc, char **argv, int type);
 void mmove(int argc, char **argv, int type);
-void mtest(int argc, char **argv, int type);
+void mpartition(int argc, char **argv, int type);
+void mtoolstest(int argc, char **argv, int type);
+void mzip(int argc, char **argv, int type);
+
+void init_privs(void);
+void reclaim_privs(void);
+void drop_privs(void);
+void destroy_privs(void);
+
+extern const char *progname;
+
+Stream_t *subdir(Stream_t *, char *pathname);
+void precmd(struct device *dev);
+
+void print_sector(char *message, unsigned char *data);
 
 #endif
