@@ -5,12 +5,30 @@
 
 #include "config.h"
 
-#ifndef __GNUC__
-#define __const const
-#define UNUSED(x) /**/
-#else
-#define UNUSED(x) x __attribute__ ((unused))
+#ifdef linux_gnu
+/* RMS strikes again */
+#ifndef linux
+#define linux
 #endif
+#endif
+
+#if defined __GNUC__ && defined __STDC__
+/* gcc -traditional doesn't have PACKED, UNUSED and NORETURN */
+#define PACKED __attribute__ ((packed))
+# if __GNUC__ == 2 && __GNUC_MINOR__ > 6 || __GNUC__ >= 3
+/* gcc 2.6.3 doesn't have "unused" */		/* mool */
+#  define UNUSED(x) x __attribute__ ((unused));x
+# else
+#  define UNUSED(x) x
+# endif
+#define NORETURN __attribute__ ((noreturn))
+#else
+#define UNUSED(x) x
+#define PACKED /* */
+#define NORETURN /* */
+
+#endif
+
 
 #include <sys/types.h>
 
@@ -25,6 +43,10 @@
 #include <unistd.h>
 #endif
 
+#ifdef HAVE_LIBC_H
+#include <libc.h>
+#endif
+
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
 #else
@@ -68,14 +90,6 @@ extern int optind, opterr;
 #endif
 
 #ifndef NO_TERMIO
-#ifdef HAVE_TERMIOS_H
-#include <termios.h>
-#endif
-
-#ifdef HAVE_SYS_TERMIOS_H
-#include <sys/termios.h>
-#endif
-
 #ifdef HAVE_TERMIO_H
 #include <termio.h>
 #endif
@@ -84,6 +98,14 @@ extern int optind, opterr;
 #ifdef HAVE_SYS_TERMIO_H
 #include <sys/termio.h>
 #endif
+
+#ifdef HAVE_TERMIOS_H
+#include <termios.h>
+#endif
+
+#ifdef HAVE_SYS_TERMIOS_H
+#include <sys/termios.h>
+#endif
 #endif
 
 #ifdef HAVE_SYS_PARAM_H
@@ -94,7 +116,10 @@ extern int optind, opterr;
 
 #include <errno.h>
 extern int errno;
+#if !defined netbsd && !defined freebsd
+/* NetBSD seems to choke on this, due to a slightly non-standard definition */
 extern char *sys_errlist[];
+#endif
 #include <pwd.h>
 
 /* On AIX, we have to prefer strings.h, as string.h lacks a prototype 
@@ -120,10 +145,15 @@ extern char *sys_errlist[];
 #include <malloc.h>
 #endif
 
+#ifdef HAVE_SIGNAL_H
+#include <signal.h>
+#else
+
 #ifdef HAVE_SYS_SIGNAL_H
 #include <sys/signal.h>
 #endif
 
+#endif
 
 #ifdef HAVE_UTIME_H
 #include <utime.h>
@@ -144,7 +174,7 @@ extern char *sys_errlist[];
 #endif
 
 #ifndef HAVE_RANDOM
-#define random lrand48
+#define random (long)lrand48
 #endif
 
 #ifndef HAVE_STRCHR
@@ -159,12 +189,12 @@ extern char *sys_errlist[];
 #define SIG_CAST RETSIGTYPE(*)()
 
 #ifndef HAVE_STRDUP
-extern char *strdup(__const char *str);
+extern char *strdup(const char *str);
 #endif /* HAVE_STRDUP */
 
 
 #ifndef HAVE_MEMCPY
-extern char *memcpy(char *s1, __const char *s2, size_t n);
+extern char *memcpy(char *s1, const char *s2, size_t n);
 #endif
 
 #ifndef HAVE_MEMSET
@@ -173,7 +203,7 @@ extern char *memset(char *s, char c, size_t n);
 
 
 #ifndef HAVE_STRPBRK
-extern char *strpbrk(__const char *string, __const char *brkset);
+extern char *strpbrk(const char *string, const char *brkset);
 #endif /* HAVE_STRPBRK */
 
 
@@ -182,65 +212,76 @@ unsigned long strtoul(const char *string, char **eptr, int base);
 #endif /* HAVE_STRTOUL */
 
 #ifndef HAVE_STRSPN
-size_t strspn(__const char *s, __const char *accept);
+size_t strspn(const char *s, const char *accept);
 #endif /* HAVE_STRSPN */
 
 #ifndef HAVE_STRCSPN
-size_t strcspn(__const char *s, __const char *reject);
+size_t strcspn(const char *s, const char *reject);
 #endif /* HAVE_STRCSPN */
 
 #ifndef HAVE_STRERROR
 char *strerror(int errno);
 #endif
 
+#ifndef HAVE_ATEXIT
+int atexit(void (*function)(void)); 
+
+#ifndef HAVE_ON_EXIT
+void myexit(int code) NORETURN;
+#define exit myexit
+#endif
+
+#endif
+
+
+#ifndef HAVE_MEMMOVE
+#define memmove(DST, SRC, N) bcopy(SRC, DST, N)
+#endif
+
 #ifndef HAVE_STRCASECMP
-int strcasecmp(const char *, const char *);
+int strcasecmp(const char *s1, const char *s2);
 #endif
+
 #ifndef HAVE_STRNCASECMP
-int strncasecmp(const char *, const char *, int);
+#ifdef __BEOS__
+int strncasecmp(const char *s1, const char *s2, unsigned int n);
+#else
+int strncasecmp(const char *s1, const char *s2, size_t n);
+#endif
 #endif
 
-#ifndef linux
-#undef USE_XDF
-#endif
-
-#ifdef NO_XDF
-#undef USE_XDF
+#ifndef HAVE_GETPASS
+char *getpass(const char *prompt);
 #endif
 
 
 #ifndef __STDC__
+#ifndef signed
 #define signed /**/
-int fflush(FILE *);
-
-#ifdef HAVE_STRDUP
-char *strdup(char *);
-#endif
-
-#ifdef HAVE_STRCASECMP
-int strcasecmp(/* const char *, const char * */);
-#endif
-#ifdef HAVE_STRNCASECMP
-int strncasecmp(/* const char *, const char *, int */);
-#endif
-char *getenv(char *);
-#ifdef HAVE_STRTOUL
-unsigned long strtoul(const char *, char **, int);
-#endif
-int pclose(FILE *);
-#ifdef HAVE_RANDOM
-long int random(void);
-#endif
-
-#ifdef HAVE_SRANDOM
-void srandom(unsigned int);
-#endif
-
-int atoi(char *);
-FILE *fdopen(int, const char *);
-FILE *popen(const char *, const char *);
+#endif 
 #endif /* !__STDC__ */
 
+/* prototypes which might be missing on some platforms, even if the functions
+ * are present.  Do not declare argument types, in order to avoid conflict
+ * on platforms where the prototypes _are_ correct.  Indeed, for most of
+ * these, there are _several_ "correct" parameter definitions, and not all
+ * platforms use the same.  For instance, some use the const attribute for
+ * strings not modified by the function, and others do not.  By using just
+ * the return type, which rarely changes, we avoid these problems.
+ */
+int fflush();
+char *strdup();
+int strcasecmp();
+int strncasecmp();
+char *getenv();
+unsigned long strtoul();
+int pclose();
+void exit();
+char *getpass();
+int atoi();
+FILE *fdopen();
+FILE *popen();
+
 #ifndef MAXPATHLEN
 #ifdef PATH_MAX
 #define MAXPATHLEN PATH_MAX
@@ -250,4 +291,14 @@ FILE *popen(const char *, const char *);
 
 #endif
 
+
+#ifndef linux
+#undef USE_XDF
+#endif
+
+#ifdef NO_XDF
+#undef USE_XDF
+#endif
+
+
 #endif
