@@ -4,7 +4,7 @@
  * written by:
  *
  * Alain L. Knaff			
- * Alain.Knaff@inrialpes.fr
+ * Alain.Knaff@poboxes.com
  *
  */
 
@@ -133,7 +133,7 @@ static int analyze_reply(RawRequest_t *raw_cmd, int do_print)
 
 
 static int send_cmd(int fd, RawRequest_t *raw_cmd, int nr,
-		    char *message, int retries)
+		    const char *message, int retries)
 {
 	int j;
 	int ret=-1;
@@ -595,11 +595,13 @@ Stream_t *XdfOpen(struct device *dev, char *name,
 	This->sector_size = 512;
 	This->stretch = 0;
 
+	precmd(dev);
 	This->fd = open(name, mode | dev->mode | O_EXCL | O_NDELAY);
 	if(This->fd < 0) {
 		sprintf(errmsg,"xdf floppy: open: \"%s\"", strerror(errno));
 		goto exit_0;
 	}
+	fcntl(This->fd, F_SETFL, FD_CLOEXEC);
 
 	This->drive = GET_DRIVE(This->fd);
 	if(This->drive < 0)
@@ -617,8 +619,8 @@ Stream_t *XdfOpen(struct device *dev, char *name,
 		goto exit_2;
 
 	/* lock the device on writes */
-	if (lock_dev(This->fd, mode == O_RDWR)) {
-		sprintf(errmsg,"xdf floppy: device \"%s\"busy\n:", 
+	if (lock_dev(This->fd, mode == O_RDWR, dev)) {
+		sprintf(errmsg,"xdf floppy: device \"%s\" busy:", 
 			dev->name);
 		goto exit_3;
 	}
